﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using TidyNetLib;
using System.IO;
using System.Reflection;
using System.Xml.Linq;
using System.Xml.XPath;
using System.Xml.Xsl;
using System.Xml;
using System.Drawing;
using System.Net;
using System.Text.RegularExpressions;
using TidyNetLib.tidy;

namespace TidyCmd
{
    /// <summary>
    /// Front end 'command line' program to open source Tidy Library 
    /// provided under MIT license. See License in Tidy.cs
    /// 
    /// TidyCmd created by Dennis Lang.
    /// Author: Dennis Lang - 2014
    /// http://home.comcast.net/~lang.dennis/    
    /// 
    /// TODO - 
    ///    * Make summary html report
    ///    * Make web link graphic showing resource links (href/url)
    ///    
    /// ========== License ==========
    /// This file is part of TidyCmd.
    ///
    /// TidyCmd is free software: you can redistribute it and/or modify
    /// it under the terms of the GNU General Public License as published by
    /// the Free Software Foundation, either version 3 of the License, or
    /// (at your option) any later version.
    ///
    /// TidyCmd is distributed in the hope that it will be useful,
    /// but WITHOUT ANY WARRANTY; without even the implied warranty of
    /// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    /// GNU General Public License for more details.
    ///
    /// The GNU General Public License: <http://www.gnu.org/licenses/>.
    /// </summary>
    class TidyCmd
    {
        Tidy mTidy;
        TidyConfig mConfig; 
        TidyMessageCollection mMessages;
        Pargs mProgArgs;

        WebPageCmd mPageTestCmd;
        WebPageCmd mPageLoadCmd;
        
        const string sEnvStr = "HTML_TIDY";
        string errMsg;

        Dictionary<MessageLevel, ConWriter.ConColors> mMsgColors = new Dictionary<MessageLevel, ConWriter.ConColors>() 
            {
                { MessageLevel.Info, new ConWriter.ConFgColor() { fg = ConsoleColor.Green }},
                { MessageLevel.Warning, new ConWriter.ConFgColor() { fg = ConsoleColor.Yellow }},
                { MessageLevel.Error, new ConWriter.ConFgColor() { fg = ConsoleColor.Magenta }},
            };

        Stream mOutput;
        ConWriter mMsgOut; 

        public TidyCmd(string[] args)
        {
            // Initialize variables

            mTidy = new Tidy();
            mConfig = new TidyConfig();
            mMessages = new TidyMessageCollection();
            mOutput = Console.OpenStandardOutput();
            mMsgOut = new ConWriter();
            mPageTestCmd = new WebPageCmd();
            mPageLoadCmd = new WebPageCmd();

            // Load default configuration set in environment variable 
            string cfgfil = Environment.GetEnvironmentVariable(sEnvStr);
            if (!string.IsNullOrEmpty(cfgfil))
            {
                if (!mConfig.LoadConfig(mTidy, cfgfil, out errMsg))
                {
                    Log.WriteError(string.Format("Loading config file {0} failed, err = {1}\n", cfgfil, errMsg));
                    Environment.Exit(-1);
                }
            }

            // Parse default and user parameter arguments.
            mProgArgs = new Pargs();
            if (!ParseArgs(args, mProgArgs))
            {
                Environment.Exit(-1);
            }
          
            if (!mProgArgs.modify && !string.IsNullOrEmpty(mProgArgs.outFile))
                mOutput = new FileStream(mProgArgs.outFile, FileMode.Create, FileAccess.Write);

            if (!string.IsNullOrEmpty(mProgArgs.errFile))
                mMsgOut.CreateFile(mProgArgs.errFile);

            if (mProgArgs.filePats.Count() != 0)
            {
                foreach (string pattern in mProgArgs.filePats)
                {
                    if (File.Exists(pattern))
                    {
                        TidyFile(pattern);   
                    }
                    else
                    {
                        // Split directory and file patterns
                        string dirPat = Path.GetDirectoryName(pattern);
                        string filePat = Path.GetFileName(pattern);
                        string[] dirs = dirPat.Split('\\');

                        GetPath(dirs, 0, filePat);
                    }
                }

            }
            else if (mProgArgs.rawArgs.Count() != 0)
            {
                foreach (List<string> inFiles in mProgArgs.rawArgs.Values)
                    foreach (string inFile in inFiles)
                        TidyFile(inFile);
            }
            else
            {
                mMsgOut.WriteLine(mMsgColors[MessageLevel.Error], "No input files specified, see -help");
                // TidyFile(null);
            }

            mTidy.SummaryReport(mMsgOut.Stream);
        }

        private void GetFiles(string pathPat)
        {
            string path = Path.GetDirectoryName(pathPat);
            string filePat = Path.GetFileName(pathPat);
            if (string.IsNullOrEmpty(path))
                path = @".\";

            foreach (string file in Directory.GetFiles(path, filePat))
            {
                FileInfo fileInfo = new FileInfo(file);
                System.Console.Error.WriteLine("Tidy:" + fileInfo.FullName);
                TidyFile(file);
            }
        }

        private void GetPath(string[] dirs, int dirIdx, string filePat)
        {
            for (; dirIdx != dirs.Length; dirIdx++)
            {
                if (Directory.Exists(dirs[dirIdx]))
                {
                    DirectoryInfo dirInfo = new DirectoryInfo(dirs[dirIdx]);
                    Environment.CurrentDirectory = dirInfo.FullName;
                }
                else
                {
                    foreach (string dir in Directory.GetDirectories(@".\", dirs[dirIdx]))
                    {
                        string currentDir = Environment.CurrentDirectory;
                        DirectoryInfo dirInfo = new DirectoryInfo(dir);

                        Environment.CurrentDirectory = dirInfo.FullName;
                        GetPath(dirs, dirIdx + 1, filePat);
                        Environment.CurrentDirectory = currentDir;
                    }
                }
            }

            GetFiles(filePat);
        }

        /// <summary>
        /// Convert relative file path to full path.
        /// </summary>
        /// <param name="inFile"></param>
        /// <returns></returns>
        private string FullPath(string inFile)
        {
            try
            {
                FileInfo fileInfo = new FileInfo(inFile);
                return fileInfo.FullName;
            }
            catch (Exception ex)
            {
            }

            return inFile;
        }

        /// <summary>
        /// Process(Tidy) an input file, optionally doing self modification and logging messages.
        /// </summary>
        /// <param name="inFile"></param>
        /// <param name="output"></param>
        /// <param name="errOut"></param>
        void TidyFile(string inFile)
        {
            inFile = FullPath(inFile);
            if (mProgArgs.modify)
            {
                mProgArgs.outFile = inFile + ".tmp";
                mOutput = new FileStream(mProgArgs.outFile, FileMode.Create, FileAccess.Write);
            }

            // Stream input = null;        //  Console.OpenStandardInput(512);
            // if (!string.IsNullOrEmpty(inFile))
            //    input = new FileStream(inFile, FileMode.Open, FileAccess.Read);
            // TidyStream(input, inFile);
            string msg = string.Empty;
            WebCmd.WebInfo webInfo = LoadPage(inFile, inFile, ref msg);
            if (!string.IsNullOrEmpty(msg))
                mMsgOut.WriteLine(mMsgColors[MessageLevel.Error], msg);
        }

        string fileTitleFmt = "----Tidy Messages for {0}----";
        private bool RealError(bool okay, string errMsg)
        {
            // Duplicate is not an error.
            return !okay && !errMsg.Equals(WebPageCmd.IgnoredDuplicate) && !errMsg.Equals(WebCmd.HostChanged);
        }

        private int TidyStream(Stream input, string inFile)
        {
            mMessages.Clear();
            mTidy.Options.Clear();  // Clear image/href/style list.
            mTidy.Parse(input, mOutput, mMessages);

            try
            {
                // Normally closed by Tidy.Parse()
                input.Close();
            }
            catch (Exception) { }

            if (mProgArgs.Showing(Pargs.Show.eProgress))
                System.Console.Error.WriteLine(string.Format(fileTitleFmt, inFile));

            int retMsgCnt = mMessages.Count;
            if (mMessages.Count != 0)
            {
                string msg = string.Format(fileTitleFmt, inFile);
                mMsgOut.WriteLine(msg);

                foreach (TidyMessage message in mMessages)
                {
                    // if (!mProgArgs.showOnlyErrors || message.Level == MessageLevel.Error)
                    if (mProgArgs.Showing(message.Level))
                    {
                        msg = message.ToString();
                        mMsgOut.WriteLine(mMsgColors[message.Level], msg);
                    }
                }
            }

            if (mProgArgs.modify)
            {
                mOutput.Close();
                mOutput = null;
                File.Delete(inFile);
                File.Move(mProgArgs.outFile, inFile);
            }

            if (!string.IsNullOrEmpty(mProgArgs.testUrls))
            {
                // TODO - handle relative, domain, etc link paths.
                // TODO - Add links (css), validate paths
                if (mTidy.Options.ImgSrcSet.Count != 0)
                {
                    if (mProgArgs.Showing(Pargs.Show.eResources))
                        mMsgOut.WriteLine("----<Img>----");
                    foreach (string imgPath in mTidy.Options.ImgSrcSet)
                    {
                        string errMsg = string.Empty;
                        WebCmd.WebInfo webInfo = IsValidImage(inFile, imgPath, ref errMsg);

                        string showPath = imgPath;
                        if (PathType(inFile, imgPath) == PathTypes.eData)
                            showPath = imgPath.Substring(0, imgPath.IndexOf(','));

                        if (RealError(webInfo, errMsg))
                            mMsgOut.WriteLine(mMsgColors[MessageLevel.Error], "Unreachable image:" + showPath + " " + errMsg);
                        else if (mProgArgs.Showing(Pargs.Show.eResources))
                            mMsgOut.WriteLine(mMsgColors[MessageLevel.Info], showPath + " " + errMsg + " " + webInfo.Summary());
                    }
                }

                if (mTidy.Options.AHrefSet.Count != 0)
                {
                    if (mProgArgs.Showing(Pargs.Show.eResources))
                        mMsgOut.WriteLine("----<A href>----");
                    foreach (string linkPath in mTidy.Options.AHrefSet)
                    {
                        string errMsg = string.Empty;
                        bool okay = false;
                        if (linkPath.StartsWith("#"))
                            okay = mTidy.Options.ANameSet.Contains(linkPath.Substring(1));
                        else
                            okay = IsValidPage(inFile, linkPath, ref errMsg);

                        if (RealError(okay, errMsg))
                            mMsgOut.WriteLine(mMsgColors[MessageLevel.Error], "Unreachable page:" + linkPath + " " + errMsg);
                        else if (mProgArgs.Showing(Pargs.Show.eResources))
                            mMsgOut.WriteLine(mMsgColors[MessageLevel.Info], linkPath + " " + errMsg);
                    }
                }

                mMsgOut.WriteLine("");
            }
            else if (mProgArgs.Showing(Pargs.Show.eResources))
            {
                if (mTidy.Options.ImgSrcSet.Count != 0)
                {
                    mMsgOut.WriteLine("----<Img>----");
                    string errMsg = string.Empty;
                    bool exists = false;
                    foreach (string imgPath in mTidy.Options.ImgSrcSet)
                    {
                        switch (PathType(inFile, imgPath, ref exists, ref errMsg))
                        {
                            case PathTypes.eFile:
                            case PathTypes.eWeb:
                                mMsgOut.WriteLine(mMsgColors[MessageLevel.Info], imgPath);
                                break;
                            case PathTypes.eUnknown:
                                mMsgOut.WriteLine(mMsgColors[MessageLevel.Info], imgPath);
                                break;
                            case PathTypes.eData:
                                mMsgOut.WriteLine(mMsgColors[MessageLevel.Info], imgPath.Substring(0, imgPath.IndexOf(',')));
                                break;
                        }
                    }
                }

                if (mTidy.Options.AHrefSet.Count != 0)
                {
                    mMsgOut.WriteLine("----<A href>----");
                    foreach (string linkPath in mTidy.Options.AHrefSet)
                    {
                        mMsgOut.WriteLine(mMsgColors[MessageLevel.Info], linkPath);
                    }
                }

                if (mTidy.Options.ANameSet.Count != 0)
                {
                    mMsgOut.WriteLine("----<A names>----");
                    foreach (string linkPath in mTidy.Options.ANameSet)
                    {
                        mMsgOut.WriteLine(mMsgColors[MessageLevel.Info], linkPath);
                    }
                }

                mMsgOut.WriteLine("");
            }

            if (mProgArgs.Showing(Pargs.Show.eResources))
            {
                if (mTidy.Options.Styles.Count != 0)
                {
                    mMsgOut.WriteLine("----<Styles>----");
                    foreach (string styleTag in this.mTidy.Options.Styles.Keys)
                    {
                        int cnt = this.mTidy.Options.Styles[styleTag];
                        if (cnt == 0)
                            mMsgOut.WriteLine(mMsgColors[MessageLevel.Warning],  string.Format("{0,20} Not Explicitly Referenced", styleTag));
                        else
                            mMsgOut.WriteLine(mMsgColors[MessageLevel.Info], string.Format("{0,20} Used:{1}", styleTag, cnt));
                    }
                }

                mMsgOut.WriteLine("");
                mTidy.NodeStats(mMsgOut.Stream);

                mMsgOut.WriteLine("");
            }

            mTidy.Summarize(inFile, mMessages);

            input.Close();
            if (!string.IsNullOrEmpty(mProgArgs.loadHrefs))
            {
                string basePath = Path.GetDirectoryName(inFile);

                var hrefSet = new HashSet<string>(mTidy.Options.AHrefSet);
                mMessages.Clear();
                foreach (string linkPath in hrefSet)
                {
                    if (!linkPath.StartsWith("#"))
                    {
                        string errMsg = string.Empty;
                        bool okay = LoadPage(inFile, linkPath, ref errMsg);
                        if (RealError(okay, errMsg))
                        {
                            if (okay)
                                mMsgOut.WriteLine(mMsgColors[MessageLevel.Error], linkPath + " " + errMsg);
                            else if (mProgArgs.Showing(Pargs.Show.eStd))
                                mMsgOut.WriteLine(mMsgColors[MessageLevel.Info], linkPath + " " + errMsg);
                        }
                    }
                }
            }

            return retMsgCnt;
        }

    
        /// <summary>
        /// Determine file path type.
        /// </summary>
        /// <param name="basePath"></param>
        /// <param name="filePath"></param>
        /// <param name="exists"></param>
        /// <param name="msg"></param>
        /// <returns>enumerated path type eFile, eWeb or eData</returns>
        Regex removeName = new Regex(@"#\w+$"); // Remove html #name tag ex: foo.html#top
        string mLocalPath;
        enum PathTypes { eUnknown, eFile, eWeb, eData };
        PathTypes PathType(string basePath, string filePath, ref bool exists, ref string msg)
        {
            mLocalPath = filePath;

            if (mLocalPath.StartsWith("data:"))
                return PathTypes.eData;

            Uri pathUri;
            if (Uri.TryCreate(mLocalPath, UriKind.RelativeOrAbsolute, out pathUri))
            {
                if (pathUri.IsAbsoluteUri && pathUri.Scheme.Equals("data"))
                    return PathTypes.eData;

                if (pathUri.IsAbsoluteUri && !string.IsNullOrEmpty(pathUri.Host))
                {
                    mLocalPath = pathUri.ToString();
                    return PathTypes.eWeb;
                }
                // if (pathUri.IsFile)
                //     return PathTypes.eFile;
            }

            mLocalPath = removeName.Replace(mLocalPath, "");
            bool isFile = mLocalPath.IndexOfAny(Path.GetInvalidPathChars()) < 0;
           
            if (isFile)
            {
                try
                {
                    if (basePath.Equals(mLocalPath))
                    {
                        exists = File.Exists(mLocalPath);
                    }
                    else
                    {
                        Uri baseUri = new Uri(basePath);
                        pathUri = new Uri(baseUri, mLocalPath);

                        if (pathUri.IsAbsoluteUri && !string.IsNullOrEmpty(pathUri.Host))
                        {
                            mLocalPath = pathUri.ToString();
                            return PathTypes.eWeb;
                        }

                        mLocalPath = pathUri.LocalPath;
                    }

                    exists = File.Exists(mLocalPath);
                    if (!exists)
                        msg = "File not found";

                    return PathTypes.eFile;
                }
                catch (Exception ex) 
                {
                    msg = ex.Message;
                }

                exists = File.Exists(mLocalPath);
                return PathTypes.eFile;
            }

            return PathTypes.eUnknown;
        }
        PathTypes PathType(string basePath, string filePath)
        {
            bool exists = false;
            string errMsg = string.Empty;
            return PathType(basePath, filePath, ref exists, ref errMsg);
        }

        /// <summary>
        /// Return true if imgPath is valid local image or web image.
        /// </summary>
        /// <param name="imgPath"></param>
        /// <param name="inFile"></param>
        /// <returns></returns>
        WebCmd.WebInfo IsValidImage(string basePath, string imgPath, ref string msg)
        {
            bool exists = false;
            switch (PathType(basePath, imgPath, ref exists, ref msg))
            {
            case PathTypes.eFile:
                try
                {
                    if (exists)
                    {
                        Bitmap.FromFile(mLocalPath);
                        return new WebCmd.WebInfo() { mValid = true, mPath = imgPath, contentType = "image",
                           contentLength = FileLength(mLocalPath)};
                    }
                }
                catch (Exception ex)
                {
                    msg = ex.Message;
                   
                }
                return null;;

            case PathTypes.eData:
                // TODO - decode base64 encoded image 
                return new WebCmd.WebInfo() { mValid = true, mPath = imgPath, contentType = "data" };

            case PathTypes.eWeb:
                return mPageTestCmd.LoadPage(basePath, imgPath, ref msg, mProgArgs.testUrls, "image", null);
            }
            return null;
        }

     
        /// <summary>
        /// Verify url can be loaded. Does not validate content.  
        /// </summary>
        /// <param name="basePath"></param>
        /// <param name="urlPath"></param>
        /// <param name="msg"></param>
        /// <returns></returns>
        bool IsValidPage(string basePath, string urlPath, ref string msg)
        {
            bool okay = false;
            switch (PathType(basePath, urlPath, ref okay, ref msg))
            {
            case PathTypes.eFile:
                break;
            case PathTypes.eWeb:
                okay = (bool)mPageTestCmd.LoadPage(basePath, urlPath, ref msg, mProgArgs.testUrls, "", null);
                break;
            }
            return okay;
        }

        /// <summary>
        /// Load page if same host and validate all pages on site.
        /// </summary>
        /// <param name="basePath"></param>
        /// <param name="urlPath"></param>
        /// <param name="msg"></param>
        /// <returns></returns>
        public WebCmd.WebInfo  LoadPage(string basePath, string urlPath, ref string msg)
        {
            bool exists = false;
            switch (PathType(basePath, urlPath, ref exists, ref msg))
            {
            case PathTypes.eFile:
                try
                {
                    if (exists && mPageLoadCmd.ValidPath(basePath, mLocalPath, mProgArgs.loadHrefs, ref msg))
                    {
                        Stream pageInStream = new FileStream(mLocalPath, FileMode.Open, FileAccess.Read);
                        TidyStream(pageInStream, mLocalPath);
                        // pageInStream.Close();
                        return new WebCmd.WebInfo() { mValid = true, mPath = mLocalPath, 
                            contentType = "text", contentLength = FileLength(mLocalPath) };
                    }
                }
                catch (Exception ex2)
                {
                    msg = ex2.Message;
                }
                return null;

            case PathTypes.eWeb:
                {
                    WebCmd.WebInfo webInfo = 
                        mPageLoadCmd.LoadPage(basePath, urlPath, ref msg, mProgArgs.loadHrefs, "text/html", TidyStream);
                    if (webInfo && webInfo.contentLength != 0)
                        mTidy.ContentLength = webInfo.contentLength;
                    return webInfo;
                }
            }

            return null;
        }


        public class Pargs
        {
            public Dictionary<string, List<string>> rawArgs = new Dictionary<string, List<string>>();
            public List<string> filePats = new List<string>();
            public string loadHrefs = null;     // Load relative href and continue Validation.
            public string testUrls = null;      // Test Image and Href urls to test they are valid
            public Arguments arguments;
            public string outFile = null;
            public string errFile = null;
            public bool  modify = false;        // Modify the original input files 

            public enum Show { eNone=0, eInfo=1, eWarn=2, eError=4, eResources=8, eStd=7, eProgress=16 };
            public int show = (int)Show.eStd;

            public bool ShowingDef { get { return show == (int)Show.eStd; } }
            public bool Showing(Show flag)
            { return (this.show & (int)flag) != 0; }
            public bool Showing(int flag)
            { return (this.show & (int)flag) != 0; }
            public bool Showing(MessageLevel msgLevel)
            {
                switch (msgLevel)
                {
                    default:
                    case MessageLevel.Info: return Showing(Show.eInfo);
                    case MessageLevel.Warning: return Showing(Show.eWarn);
                    case MessageLevel.Error: return Showing(Show.eError);
                }
            }

            public void SetShowing(Show flag) 
            {
                if (ShowingDef)
                    show = (int)flag;
                else
                    show |= (int)flag;
            }
        }

        public bool ParseArgs(string[] args, Pargs pargs)
        {
            if (args.Length == 0)
            {
                TidyHelp.ShowHelp(TidyHelp.HelpLevel.eHelpCmd);
                return false;
            }

            pargs.arguments = new Arguments(args, false, ref pargs.rawArgs);

            bool okay = true;
            foreach (string cmd in pargs.arguments.Keys)
            {
                string value = pargs.arguments[cmd];
                // if (pargs.arguments.TryGetValueAndRemove(cmd, out value))
                {
                    string errMsg = string.Empty;
                    int num = 0;
                    
                    switch (cmd.ToLower())
                    {
                        // ##### HELP #####
                        case "help":
                        case "h":
                        case "?":
                            TidyHelp.ShowHelp(TidyHelp.HelpLevel.eHelpCmd);
                            return false;
                        case "helpfull":
                            TidyHelp.ShowHelp(TidyHelp.HelpLevel.eHelpFull);
                            return false;

                        case "help-xml":        // List the command line options in XML format 
                            TidyHelp.ShowHelp(TidyHelp.HelpLevel.eHelpXml);
                            return false;

                        case "help-config":     // List all configuration options 
                            TidyHelp.ShowHelp(TidyHelp.HelpLevel.eHelpCfg);
                            return false;

                        case "xml-config":      // List all configuration options in XML format 
                            throw new NotImplementedException();

                        case "show-config":     // List the current configuration settings
                            mConfig.ShowConfig(mTidy, Console.Out);
                            return false;

                        case "version":
                        case "v":
                            TidyHelp.ShowHelp(TidyHelp.HelpLevel.eHelpVersion);
                            return false;

                        // ##### FILE MANIPULATION ######
                        case "file":    // Redirect errors and warnings to file.
                        case "logfile":
                            Log.FilePath = value;
                            value = string.Empty;
                            break;


                        case "input":
                        case "infile":
                        case "filepats":    // File patterns,  ex: dir\dir1\*.html  or  dir\*\*.html
                            pargs.filePats.Add(value);
                            value = string.Empty;
                            break;
             
                        case "output":
                        case "outfile":
                            pargs.outFile = value;
                            value = string.Empty;
                            break;

                        case "errfile":
                            pargs.errFile = value;
                            value = string.Empty;
                            break;

                        case "config":  // Load configuration options from file
                            if (!mConfig.LoadConfig(mTidy, value, out errMsg))
                            {
                                Log.WriteError(errMsg);
                                okay = false;
                            }
                            value = string.Empty;
                            break;

                        case "modify":  // Modify input file.
                            // mTidy.Options.WriteBack = true;
                            pargs.modify = true;
                            break;

                        // ##### VALIDATION ######
                        case "errors":  // Show only errors, defaults to Info+Warn+Error
                            pargs.SetShowing(Pargs.Show.eError);
                            break;

                        case "warnings":  // Show warnings. 
                            pargs.SetShowing(Pargs.Show.eWarn);
                            break;

                        case "resources":  // Show image and href resource references
                            pargs.show |= (int)Pargs.Show.eResources;
                            break;

                        case "testurls":   // Test Image and Links (relative|...) to validate url
                            pargs.testUrls = string.IsNullOrEmpty(value) ? "relative" : value.ToLower();
                            value = string.Empty;
                            break;

                        case "loadhrefs":   // Load href links (relative|...) and continue validation. 
                            pargs.loadHrefs = string.IsNullOrEmpty(value) ? "relative" : value.ToLower();
                            value = string.Empty;
                            break;

                        case "progress":
                            pargs.show |= (int)Pargs.Show.eProgress;
                            break;

                        // ##### FORMAT OUTOUT #####
                        case "indent":  // Enable indentation
                            mTidy.Options.IndentContent = true;
                            break;

                        case "wrap":    // Set wrap length
                            if (!int.TryParse(value, out num))
                            {
                                Log.WriteError("Failed to parse numeric wrap length from:" + value);
                                okay = false;
                            }
                            else
                                mTidy.Options.WrapLen = num;
                            value = string.Empty;
                            break;

                        case "upper":   // Force tags to upper case
                            mTidy.Options.UpperCaseTags = true;
                            break;

                        case "clean":   // Replace FONT, NOBR and CENTER tags by CSS
                            mTidy.Options.MakeClean = true;
                            break;

                        case "numeric": // Output numeric rather than named entities
                            mTidy.Options.NumEntities = true;
                            break;

                        case "omit":    // Omit optional start tags and end tags 
                            mTidy.Options.HideEndTags = true;
                            break;

                        case "xml":     // Specify the input is well formed XML
                            mTidy.Options.XmlTags = true;
                            break;

                        case "asxml":   // Convert HTML to well formed xml
                            mTidy.Options.XmlOut = true;
                            break;
                        case "asxhtml": // Convert HTML to well formed XHTML
                            mTidy.Options.XhtmlOut = true;
                            break;
                        case "ashtml":  // Force XHTML to well formed HTML
                            mTidy.Options.XhtmlOut = true;
                            break;
 
                        case "raw":     // Output values above 127 without conversion to entities 
                        case "ascii":   // Use ISO-8859-1 for input, US-ASCII for output 
                        case "latin0":  // Use ISO-8859-15 for input, US-ASCII for output 
                        case "latin1":  // Use ISO-8859-1 for both input and output 
                        case "utf8":    // Use UTF-8 for both input and output 
                            // throw new NotImplementedException();
                            if (!mConfig.Parse(mTidy, TidyConfig.GetOptionName(TidyOptionId.eInCharEncoding), cmd))
                            {
                                errMsg += "Unknown encoding:" + cmd;
                            }
                            break;

                        // ######  [NOT IMPLEMENTED YET]  #######
                        case "access":  // Do additional accessibility checks (<%0blevel%0f> = 0, 1, 2, 3).
                            throw new NotImplementedException();
                        case "bare":    // Strip out smart quotes and em dashes, etc.
                            // mTidy.Options.FixBackslash
                            throw new NotImplementedException();
                        case "quiet":    // Suppress nonessential output
                            throw new NotImplementedException();
                        case "gdoc":    // Produce clean version of html exported by google docs
                            throw new NotImplementedException();
                        case "xfilepats":   // exclude
                            throw new NotImplementedException();

                        default:
                            if (!mConfig.Parse(mTidy, cmd, value))
                            {
                                errMsg = "Unknown option or value:" + cmd + " " + value;
                                mConfig.GetPossibleValues(cmd, ref errMsg);
                            }
                            break;
                    }

                    if (!string.IsNullOrEmpty(errMsg))
                    {
                        Console.Error.WriteLine(errMsg);
                        errMsg = string.Empty;
                        okay = false;
                    }
                }

                // Transfer an unused arguments into raw list.
                if (!string.IsNullOrEmpty(value) && !value.Equals(Arguments.sTrue))
                {
                    if (!pargs.rawArgs.ContainsKey(""))
                        pargs.rawArgs.Add("", new List<string>());
                    pargs.rawArgs[""].Add(value);
                }
            }

            return okay;
        }


        static long FileLength(string path)
        {
            try
            {
                FileInfo fileInfo = new FileInfo(path);
                return fileInfo.Length;
            }
            catch { }
            return 0;
        }
    }

}
