﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using TidyNetLib;
using System.Reflection;
using System.IO;

namespace TidyCmd
{
    /// <summary>
    /// TidyOptionImpl - abstract base class to command option parsing classes.
    /// 
    /// Author: Dennis Lang - 2014
    /// http://home.comcast.net/~lang.dennis/    
    /// 
    /// ========== License ==========
    /// This file is part of TidyCmd.
    ///
    /// TidyCmd is free software: you can redistribute it and/or modify
    /// it under the terms of the GNU General Public License as published by
    /// the Free Software Foundation, either version 3 of the License, or
    /// (at your option) any later version.
    ///
    /// TidyCmd is distributed in the hope that it will be useful,
    /// but WITHOUT ANY WARRANTY; without even the implied warranty of
    /// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    /// GNU General Public License for more details.
    ///
    /// The GNU General Public License: <http://www.gnu.org/licenses/>.
    /// </summary>
    public abstract class TidyOptionImpl
    {
        public TidyOptionId m_id;               // Enum matches TidyOptions property names
        public TidyConfigCategory m_category;   // put 'em in groups 
        public string m_name;                   // property name 
        public TidyOptionType m_type;           // string, int or bool  
        public string[] m_pickList;             // pick list  

        public TidyOptionImpl(
            TidyOptionId        id,
            TidyConfigCategory  category, 
            string              name,
            TidyOptionType      type,
            string[]            pickList)
        {
            m_id            =  id;
            m_category      =  category; 
            m_name          =  name;     
            m_type          =  type;     
            m_pickList      =  pickList; 
        }

        public abstract bool Parse(Tidy updTidy, string inStr);
        public abstract string DefaultStr { get; }

        public bool GetPickList<TT>(string inStr, ref TT outVal)
        {
            if (!string.IsNullOrEmpty(inStr))
            {
                string inLwr = inStr.ToLower();

                for (int idx = 0; idx != m_pickList.Length; idx++)
                {
                    string[] picks = m_pickList[idx].Split(',');
                    foreach (string pickStr in picks)
                    {
                        if (pickStr.Trim().ToLower().StartsWith(inLwr))
                        {
                            outVal = (TT)Enum.GetValues(typeof(TT)).GetValue(idx);
                            return true;
                        }
                    }
                }
            }
            return false;
        }

        public static bool GetBool(string inStr, ref bool bValue)
        {
            if (!string.IsNullOrEmpty(inStr))
            {
                switch (char.ToLower(inStr[0]))
                {
                    case 't':   // true
                    case 'y':   // yes
                        bValue = true;
                        return true;
                    case 'f':   // false
                    case 'n':   // no
                        bValue = false;
                        return true;
                }
            }
            return false;
        }

        public object Get(TidyOptions options)
        {
            MethodInfo getMethodInfo = TidyConfig.sOptAndPropDic[m_id].mPropertyInfo.GetGetMethod();
            return getMethodInfo.Invoke(options, null);
        }

        public void Set(TidyOptions options, object objValue)
        {
            MethodInfo setMethodInfo = TidyConfig.sOptAndPropDic[m_id].mPropertyInfo.GetSetMethod();
            setMethodInfo.Invoke(options, new object[] { objValue });
        }
    };

    /// <summary>
    /// Parse boolean(True/False) and Auto command.
    /// </summary>
    public class ParseAutoBool : TidyOptionImpl
    {
        public ParseAutoBool(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          TidyTriState dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt;  }

        public TidyTriState Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            TidyTriState bValue = Default;
            bool valid = GetPickList<TidyTriState>(inStr, ref bValue);

            Set(updTidy.Options, bValue);
            // Default = bValue;

            return valid;
        }
    }

    /// <summary>
    /// Parse boolean (True/False)
    /// </summary>
    public class ParseBool : TidyOptionImpl
    {
        public ParseBool(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          bool dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt; }

        public bool Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            bool bValue = Default;
            bool valid = GetBool(inStr, ref bValue);

            Set(updTidy.Options, bValue);
            // Default = bValue;

            return valid;
        }
    }

    /// <summary>
    /// Parse Character encoding types.
    /// </summary>
    public class ParseCharEnc : TidyOptionImpl
    {
        public ParseCharEnc(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          Encoding dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt;  }

        public Encoding Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            bool valid = !string.IsNullOrEmpty(inStr);
            Encoding enc = Default;

            if (valid)
            {
                valid = false;
                string inLwr = inStr.ToLower();
                foreach (EncodingInfo encInfo in Encoding.GetEncodings())
                {
                    if (encInfo.Name.ToLower().Equals(inLwr))
                    {
                        valid = true;
                        enc = encInfo.GetEncoding();
                        break;
                    }
                }
            }

            Set(updTidy.Options, enc);
            // Default = enc;

            return valid;
        }
    }

    public class ParseCSS1Selector : TidyOptionImpl
    {
        public ParseCSS1Selector(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          string dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt; }

        public string Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            bool valid = !string.IsNullOrEmpty(inStr);
            string value = Default;

            Set(updTidy.Options, value);
            // Default = value;

            return valid;
        }
    }

    /// <summary>
    /// Parse Document type.
    /// </summary>
    public class ParseDocType : TidyOptionImpl
    {
        public ParseDocType(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          TidyDoctypeModes dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt; }

        public TidyDoctypeModes Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            TidyDoctypeModes value = Default;
            bool valid = GetPickList<TidyDoctypeModes>(inStr, ref value);
            Set(updTidy.Options, value);
            // Default = value;

            return valid;
        }
    }

    /// <summary>
    /// Parse integer.
    /// </summary>
    public class ParseInt : TidyOptionImpl
    {
        public ParseInt(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          ulong dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt;  }

        public ulong Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            bool valid = !string.IsNullOrEmpty(inStr);
            ulong value = Default;
            valid = valid && ulong.TryParse(inStr, out value);
            Set(updTidy.Options, value);
            // Default = value;

            return valid;
        }
    }

    public class ParseName : TidyOptionImpl
    {
        public ParseName(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          string dflt,
          string[] pickList ) :
            base(id, category, name, type, pickList)
        { }

        public string Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            bool valid = !string.IsNullOrEmpty(inStr);
            string value = Default;

            if (valid)
            {
                value = inStr.ToLower();
                if (Array.IndexOf(m_pickList, value) == -1)
                {
                    value = Default;
                    valid = false;
                }
            }

            Set(updTidy.Options, value);
            // Default = value;

            return valid;
        }
    }

    public class ParseNewline : TidyOptionImpl
    {
        public ParseNewline(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          TidyLineEnding dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt;  }

        public TidyLineEnding Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            TidyLineEnding value = Default;
            bool valid = GetPickList<TidyLineEnding>(inStr, ref value);

            Set(updTidy.Options, value);
            // Default = value;

            return valid;
        }
    }

    public class ParseRepeatAttr : TidyOptionImpl
    {
        public ParseRepeatAttr(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          TidyDupAttrModes dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt;  }

        public TidyDupAttrModes Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            TidyDupAttrModes value = Default;
            bool valid = GetPickList<TidyDupAttrModes>(inStr, ref value);

            Set(updTidy.Options, value);
            // Default = value;

            return valid;
        }
    }

    public class ParseSorter : TidyOptionImpl
    {
        public ParseSorter(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          TidyAttrSortStrategy dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt;  }

        public TidyAttrSortStrategy Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            TidyAttrSortStrategy value = Default;
            bool valid = GetPickList<TidyAttrSortStrategy>(inStr, ref value);

            Set(updTidy.Options, value);
            // Default = value;

            return valid;
        }
    }

    public class ParseString : TidyOptionImpl
    {
        public ParseString(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          string dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt;  }

        public string Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            bool valid = !string.IsNullOrEmpty(inStr);
            string value = Default;

            if (valid)
            {
                value = inStr.ToLower();
                if (Array.IndexOf(m_pickList, value) == -1)
                {
                    value = Default;
                    valid = false;
                }
            }

            Set(updTidy.Options, value);
            // Default = value;

            return valid;
        }
    }

    public class ParseTagNames  : TidyOptionImpl
    {
        public ParseTagNames(
          TidyOptionId id,
          TidyConfigCategory category,
          string name,
          TidyOptionType type,
          string dflt,
          string[] pickList) :
            base(id, category, name, type, pickList)
        { Default = dflt; }

        public string Default { get; set; }
        public override string DefaultStr { get { return Default.ToString(); } }

        public override bool Parse(Tidy updTidy, string inStr)
        {
            bool valid = !string.IsNullOrEmpty(inStr);
            string value = Default;

            Set(updTidy.Options, value);
            // Default = value;

            return valid;
        }
    }

    //  TidyOptionId::TidyUnknownOption
    // static TidyOptionImpl  option_defs; //   = new TidyOptionImpl[2];
 
    /*
          { TidyUnknownOption,           MS, "unknown!",                    IN, 0,               NULL,              NULL            },
          { TidyIndentSpaces,            PP, "indent-spaces",               IN, 2,               ParseInt,          NULL            },
          { TidyWrapLen,                 PP, "wrap",                        IN, 68,              ParseInt,          NULL            },
          { TidyTabSize,                 PP, "tab-size",                    IN, 8,               ParseInt,          NULL            },
          { TidyCharEncoding,            CE, "char-encoding",               IN, UTF8,            ParseCharEnc,      charEncPicks    },
          { TidyInCharEncoding,          CE, "input-encoding",              IN, UTF8,            ParseCharEnc,      charEncPicks    },
          { TidyOutCharEncoding,         CE, "output-encoding",             IN, UTF8,            ParseCharEnc,      charEncPicks    },
          { TidyNewline,                 CE, "newline",                     IN, DLF,             ParseNewline,      newlinePicks    },
          { TidyDoctypeMode,             MU, "doctype-mode",                IN, TidyDoctypeAuto, NULL,              doctypePicks    },
          { TidyDoctype,                 MU, "doctype",                     ST, 0,               ParseDocType,      doctypePicks    },
          { TidyDuplicateAttrs,          MU, "repeated-attributes",         IN, TidyKeepLast,    ParseRepeatAttr,   repeatAttrPicks },
          { TidyAltText,                 MU, "alt-text",                    ST, 0,               ParseString,       NULL            },

          //  obsolete 
          { TidySlideStyle,              MS, "slide-style",                 ST, 0,               ParseName,         NULL            },

          { TidyErrFile,                 MS, "error-file",                  ST, 0,               ParseString,       NULL            },
          { TidyOutFile,                 MS, "output-file",                 ST, 0,               ParseString,       NULL            },
          { TidyWriteBack,               MS, "write-back",                  BL, no,              ParseBool,         boolPicks       },
          { TidyShowMarkup,              PP, "markup",                      BL, yes,             ParseBool,         boolPicks       },
          { TidyShowInfo,                DG, "show-info",                   BL, yes,             ParseBool,         boolPicks       },
          { TidyShowWarnings,            DG, "show-warnings",               BL, yes,             ParseBool,         boolPicks       },
          { TidyQuiet,                   MS, "quiet",                       BL, no,              ParseBool,         boolPicks       },
          { TidyIndentContent,           PP, "indent",                      IN, TidyNoState,     ParseAutoBool,     autoBoolPicks   },
          { TidyCoerceEndTags,           MU, "coerce-endtags",              BL, yes,             ParseBool,         boolPicks       },
          { TidyOmitOptionalTags,        MU, "omit-optional-tags",          BL, no,              ParseBool,         boolPicks       },
          { TidyHideEndTags,             MU, "hide-endtags",                BL, no,              ParseBool,         boolPicks       },
          { TidyXmlTags,                 MU, "input-xml",                   BL, no,              ParseBool,         boolPicks       },
          { TidyXmlOut,                  MU, "output-xml",                  BL, no,              ParseBool,         boolPicks       },
          { TidyXhtmlOut,                MU, "output-xhtml",                BL, no,              ParseBool,         boolPicks       },
          { TidyHtmlOut,                 MU, "output-html",                 BL, no,              ParseBool,         boolPicks       },
          { TidyXmlDecl,                 MU, "add-xml-decl",                BL, no,              ParseBool,         boolPicks       },
          { TidyUpperCaseTags,           MU, "uppercase-tags",              BL, no,              ParseBool,         boolPicks       },
          { TidyUpperCaseAttrs,          MU, "uppercase-attributes",        BL, no,              ParseBool,         boolPicks       },
          { TidyMakeBare,                MU, "bare",                        BL, no,              ParseBool,         boolPicks       },
          { TidyMakeClean,               MU, "clean",                       BL, no,              ParseBool,         boolPicks       },
          { TidyGDocClean,               MU, "gdoc",                        BL, no,              ParseBool,         boolPicks       },
          { TidyLogicalEmphasis,         MU, "logical-emphasis",            BL, no,              ParseBool,         boolPicks       },
          { TidyDropPropAttrs,           MU, "drop-proprietary-attributes", BL, no,              ParseBool,         boolPicks       },
          { TidyDropFontTags,            MU, "drop-font-tags",              BL, no,              ParseBool,         boolPicks       },
          { TidyDropEmptyElems,          MU, "drop-empty-elements",         BL, yes,             ParseBool,         boolPicks       },
          { TidyDropEmptyParas,          MU, "drop-empty-paras",            BL, yes,             ParseBool,         boolPicks       },
          { TidyFixComments,             MU, "fix-bad-comments",            BL, yes,             ParseBool,         boolPicks       },
          { TidyBreakBeforeBR,           PP, "break-before-br",             BL, no,              ParseBool,         boolPicks       },

          //  obsolete  
          { TidyBurstSlides,             PP, "split",                       BL, no,              ParseBool,         boolPicks       },

          { TidyNumEntities,             MU, "numeric-entities",            BL, no,              ParseBool,         boolPicks       },
          { TidyQuoteMarks,              MU, "quote-marks",                 BL, no,              ParseBool,         boolPicks       },
          { TidyQuoteNbsp,               MU, "quote-nbsp",                  BL, yes,             ParseBool,         boolPicks       },
          { TidyQuoteAmpersand,          MU, "quote-ampersand",             BL, yes,             ParseBool,         boolPicks       },
          { TidyWrapAttVals,             PP, "wrap-attributes",             BL, no,              ParseBool,         boolPicks       },
          { TidyWrapScriptlets,          PP, "wrap-script-literals",        BL, no,              ParseBool,         boolPicks       },
          { TidyWrapSection,             PP, "wrap-sections",               BL, yes,             ParseBool,         boolPicks       },
          { TidyWrapAsp,                 PP, "wrap-asp",                    BL, yes,             ParseBool,         boolPicks       },
          { TidyWrapJste,                PP, "wrap-jste",                   BL, yes,             ParseBool,         boolPicks       },
          { TidyWrapPhp,                 PP, "wrap-php",                    BL, yes,             ParseBool,         boolPicks       },
          { TidyFixBackslash,            MU, "fix-backslash",               BL, yes,             ParseBool,         boolPicks       },
          { TidyIndentAttributes,        PP, "indent-attributes",           BL, no,              ParseBool,         boolPicks       },
          { TidyXmlPIs,                  MU, "assume-xml-procins",          BL, no,              ParseBool,         boolPicks       },
          { TidyXmlSpace,                MU, "add-xml-space",               BL, no,              ParseBool,         boolPicks       },
          { TidyEncloseBodyText,         MU, "enclose-text",                BL, no,              ParseBool,         boolPicks       },
          { TidyEncloseBlockText,        MU, "enclose-block-text",          BL, no,              ParseBool,         boolPicks       },
          { TidyKeepFileTimes,           MS, "keep-time",                   BL, no,              ParseBool,         boolPicks       },
          { TidyWord2000,                MU, "word-2000",                   BL, no,              ParseBool,         boolPicks       },
          { TidyMark,                    MS, "tidy-mark",                   BL, yes,             ParseBool,         boolPicks       },
          { TidyEmacs,                   MS, "gnu-emacs",                   BL, no,              ParseBool,         boolPicks       },
          { TidyEmacsFile,               MS, "gnu-emacs-file",              ST, 0,               ParseString,       NULL            },
          { TidyLiteralAttribs,          MU, "literal-attributes",          BL, no,              ParseBool,         boolPicks       },
          { TidyBodyOnly,                MU, "show-body-only",              IN, no,              ParseAutoBool,     autoBoolPicks   },
          { TidyFixUri,                  MU, "fix-uri",                     BL, yes,             ParseBool,         boolPicks       },
          { TidyLowerLiterals,           MU, "lower-literals",              BL, yes,             ParseBool,         boolPicks       },
          { TidyHideComments,            MU, "hide-comments",               BL, no,              ParseBool,         boolPicks       },
          { TidyIndentCdata,             MU, "indent-cdata",                BL, no,              ParseBool,         boolPicks       },
          { TidyForceOutput,             MS, "force-output",                BL, no,              ParseBool,         boolPicks       },
          { TidyShowErrors,              DG, "show-errors",                 IN, 6,               ParseInt,          NULL            },
          { TidyAsciiChars,              CE, "ascii-chars",                 BL, no,              ParseBool,         boolPicks       },
          { TidyJoinClasses,             MU, "join-classes",                BL, no,              ParseBool,         boolPicks       },
          { TidyJoinStyles,              MU, "join-styles",                 BL, yes,             ParseBool,         boolPicks       },
          { TidyEscapeCdata,             MU, "escape-cdata",                BL, no,              ParseBool,         boolPicks       },
        #if SUPPORT_ASIAN_ENCODINGS
          { TidyLanguage,                CE, "language",                    ST, 0,               ParseName,         NULL            },
          { TidyNCR,                     MU, "ncr",                         BL, yes,             ParseBool,         boolPicks       },
        #endif
        #if SUPPORT_UTF16_ENCODINGS
          { TidyOutputBOM,               CE, "output-bom",                  IN, TidyAutoState,   ParseAutoBool,     autoBoolPicks   },
        #endif
          { TidyReplaceColor,            MU, "replace-color",               BL, no,              ParseBool,         boolPicks       },
          { TidyCSSPrefix,               MU, "css-prefix",                  ST, 0,               ParseCSS1Selector, NULL            },
          { TidyInlineTags,              MU, "new-inline-tags",             ST, 0,               ParseTagNames,     NULL            },
          { TidyBlockTags,               MU, "new-blocklevel-tags",         ST, 0,               ParseTagNames,     NULL            },
          { TidyEmptyTags,               MU, "new-empty-tags",              ST, 0,               ParseTagNames,     NULL            },
          { TidyPreTags,                 MU, "new-pre-tags",                ST, 0,               ParseTagNames,     NULL            },
          { TidyAccessibilityCheckLevel, DG, "accessibility-check",         IN, 0,               ParseAcc,          accessPicks     },
          { TidyVertSpace,               PP, "vertical-space",              BL, no,              ParseBool,         boolPicks       },
        #if SUPPORT_ASIAN_ENCODINGS
          { TidyPunctWrap,               PP, "punctuation-wrap",            BL, no,              ParseBool,         boolPicks       },
        #endif
          { TidyMergeEmphasis,           MU, "merge-emphasis",              BL, yes,             ParseBool,         boolPicks       },
          { TidyMergeDivs,               MU, "merge-divs",                  IN, TidyAutoState,   ParseAutoBool,     autoBoolPicks   },
          { TidyDecorateInferredUL,      MU, "decorate-inferred-ul",        BL, no,              ParseBool,         boolPicks       },
          { TidyPreserveEntities,        MU, "preserve-entities",           BL, no,              ParseBool,         boolPicks       },
          { TidySortAttributes,          PP, "sort-attributes",             IN, TidySortAttrNone,ParseSorter,       sorterPicks     },
          { TidyMergeSpans,              MU, "merge-spans",                 IN, TidyAutoState,   ParseAutoBool,     autoBoolPicks   },
          { TidyAnchorAsName,            MU, "anchor-as-name",              BL, yes,             ParseBool,         boolPicks       },
          { N_TIDY_OPTIONS,              XX, NULL,                          XY, 0,               NULL,              NULL            }
    );
    */

    public class TidyConfig
    {
        // Arrange so index can be cast to enum
        static string[] boolPicks = new string[] 
        {
          "no, false",
          "yes, true"
        };

        static string[] autoBoolPicks  = new string[] 
        {
          "no, false",
          "yes, true",
          "auto" 
        };

        static string[] repeatAttrPicks  = new string[] 
        {
          "keep-first, keepFirst",
          "keep-last, keepLast" 
        };

        static string[] accessPicks  = new string[] 
        {
          "0 (Tidy Classic)",
          "1 (Priority 1 Checks)",
          "2 (Priority 2 Checks)",
          "3 (Priority 3 Checks)"
        };

        static string[] charEncPicks  = new string[] 
        {
          "raw",
          "ascii",
          "latin0",
          "latin1",
          "utf8",
        //#ifndef NO_NATIVE_ISO2022_SUPPORT
        //  "iso2022",
        //#endif
          "mac",
          "win1252",
          "ibm858",

#if false
        #if SUPPORT_UTF16_ENCODINGS
          "utf16le",
          "utf16be",
          "utf16",
        #endif

        #if SUPPORT_ASIAN_ENCODINGS
          "big5",
          "shiftjis",
        #endif
#endif
        };

        static string[] newlinePicks  = new string[] 
        {
          "LF",
          "CRLF",
          "CR"
        };

        static string[] doctypePicks  = new string[] 
        {
          "html5",
          "omit",
          "auto",
          "strict",
          "transitional",
          "user"
        };

        static string[] sorterPicks  = new string[] 
        {
          "none",
          "alpha"
        };

        const TidyConfigCategory MU = TidyConfigCategory.eMarkup;
        const TidyConfigCategory DG = TidyConfigCategory.eDiagnostics;
        const TidyConfigCategory PP = TidyConfigCategory.ePrettyPrint;
        const TidyConfigCategory CE = TidyConfigCategory.eEncoding;
        const TidyConfigCategory MS = TidyConfigCategory.eMiscellaneous;

        const TidyOptionType IN = TidyOptionType.eInteger;
        const TidyOptionType BL = TidyOptionType.eBoolean;
        const TidyOptionType ST = TidyOptionType.eString;

        const string NullStr = null;
        const string[] NullArr = null;

// #define XX (TidyConfigCategory)-1
// #define XY (TidyOptionType)-1

// #define DLF DEFAULT_NL_CONFIG

        public static TidyOptionImpl[] sOptionDefaults;

        public class OptAndProp
        {
            public PropertyInfo     mPropertyInfo;
            public TidyOptionImpl   mOptionImpl;
        };

        public static Dictionary<TidyOptionId, OptAndProp> sOptAndPropDic = null;
        public static Dictionary<string, TidyOptionId> sOptToIdDic = null;
      

        public TidyConfig()
        {
            sOptionDefaults = new TidyOptionImpl[]{
            new ParseInt(TidyOptionId.eIndentSpaces,                PP, "indent-spaces",                IN, 2,              NullArr),
            new ParseInt(TidyOptionId.eWrapLen,                     PP, "wrap",                         IN, 150,            NullArr),
            new ParseInt(TidyOptionId.eTabSize,                     PP, "tab-size",                     IN, 8,              NullArr),
            new ParseCharEnc(TidyOptionId.eCharEncoding,            CE, "char-encoding",                IN, Encoding.UTF8,  charEncPicks),      
            new ParseCharEnc(TidyOptionId.eInCharEncoding,          CE, "input-encoding",               IN, Encoding.UTF8,  charEncPicks),        
            new ParseCharEnc(TidyOptionId.eOutCharEncoding,         CE, "output-encoding",              IN, Encoding.UTF8,  charEncPicks),    
  //          new ParseNewline(TidyOptionId.eNewline,                 CE, "newline",                      IN, TidyLineEnding.eCRLF,   newlinePicks),
            new ParseDocType(TidyOptionId.eDocType,                 MU, "doctype",                      ST, TidyDoctypeModes.eAuto, doctypePicks),
            new ParseRepeatAttr(TidyOptionId.eDuplicateAttrs,       MU, "repeated-attributes",          IN, TidyDupAttrModes.eKeepLast, repeatAttrPicks),
            new ParseString(TidyOptionId.eAltText,                  MU, "alt-text",                     ST, NullStr,        NullArr),
            new ParseName(TidyOptionId.eSlideStyle,                 MS, "slide-style",                  ST, NullStr,        NullArr),    // Obsolete
   //         new ParseString(TidyOptionId.eErrFile,                  MS, "error-file",                   ST, NullStr,            NullArr         ),
   //         new ParseString(TidyOptionId.eOutFile,                  MS, "output-file",                  ST, NullStr,            NullArr         ),
   //         new ParseBool(TidyOptionId.eWriteBack,                  MS, "write-back",                   BL, false,              boolPicks       ),
   //         new ParseBool(TidyOptionId.eShowMarkup,                 PP, "markup",                       BL, true,               boolPicks       ),
   //         new ParseBool(TidyOptionId.eShowInfo,                   DG, "show-info",                    BL, true,               boolPicks       ),
   //         new ParseBool(TidyOptionId.eShowWarnings,               DG, "show-warnings",                BL, true,               boolPicks       ),
   //         new ParseBool(TidyOptionId.eQuiet,                      MS, "quiet",                        BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eIndentContent,              PP, "indent",                       IN, true,               boolPicks   ),
    //?        new ParseBool(TidyOptionId.eCoerceEndTags,              MU, "coerce-endtags",               BL, true,               boolPicks       ),
    //?        new ParseBool(TidyOptionId.eOmitOptionalTags,           MU, "omit-optional-tags",           BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eHideEndTags,                MU, "hide-endtags",                 BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eXmlTags,                    MU, "input-xml",                    BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eXmlOut,                     MU, "output-xml",                   BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eXhtmlOut,                   MU, "output-xhtml",                 BL, false,              boolPicks       ),
    //        new ParseBool(TidyOptionId.eHtmlOut,                    MU, "output-html",                  BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eXmlDecl,                    MU, "add-xml-decl",                 BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eUpperCaseTags,              MU, "uppercase-tags",               BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eUpperCaseAttrs,             MU, "uppercase-attributes",         BL, false,              boolPicks       ),
    //        new ParseBool(TidyOptionId.eMakeBare,                   MU, "bare",                         BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eMakeClean,                  MU, "clean",                        BL, false,              boolPicks       ),
    //        new ParseBool(TidyOptionId.eGDocClean,                  MU, "gdoc",                         BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eLogicalEmphasis,            MU, "logical-emphasis",             BL, false,              boolPicks       ),
    //        new ParseBool(TidyOptionId.eDropPropAttrs,              MU, "drop-proprietary-attributes",  BL, false,              boolPicks       ),
            new ParseBool(TidyOptionId.eDropFontTags,               MU, "drop-font-tags",               BL, false,              boolPicks       ),
    //        new ParseBool(TidyOptionId.eDropEmptyElems,             MU, "drop-empty-elements",          BL, true,               boolPicks       ),
            new ParseBool(TidyOptionId.eDropEmptyParas,             MU, "drop-empty-paras",             BL, true,               boolPicks       ),
            new ParseBool(TidyOptionId.eFixComments,                MU, "fix-bad-comments",             BL, true,               boolPicks       ),
            new ParseBool(TidyOptionId.eBreakBeforeBR,              PP, "break-before-br",              BL, false,              boolPicks       ),

            //  obsolete  
            new ParseBool(TidyOptionId.eBurstSlides,             PP, "split",                       BL, false,          boolPicks       ),
            
            new ParseBool(TidyOptionId.eNumEntities,             MU, "numeric-entities",            BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eQuoteMarks,              MU, "quote-marks",                 BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eQuoteNbsp,               MU, "quote-nbsp",                  BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eQuoteAmpersand,          MU, "quote-ampersand",             BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eWrapAttVals,             PP, "wrap-attributes",             BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eWrapScriptlets,          PP, "wrap-script-literals",        BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eWrapSection,             PP, "wrap-sections",               BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eWrapAsp,                 PP, "wrap-asp",                    BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eWrapJste,                PP, "wrap-jste",                   BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eWrapPhp,                 PP, "wrap-php",                    BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eFixBackslash,            MU, "fix-backslash",               BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eIndentAttributes,        PP, "indent-attributes",           BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eXmlPIs,                  MU, "assume-xml-procins",          BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eXmlSpace,                MU, "add-xml-space",               BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eEncloseBodyText,         MU, "enclose-text",                BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eEncloseBlockText,        MU, "enclose-block-text",          BL, false,          boolPicks       ),
     //       new ParseBool(TidyOptionId.eKeepFileTimes,           MS, "keep-time",                   BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eWord2000,                MU, "word-2000",                   BL, false,          boolPicks       ),
     //       new ParseBool(TidyOptionId.eMark,                    MS, "tidy-mark",                   BL, true,           boolPicks       ),
     //       new ParseBool(TidyOptionId.eEmacs,                   MS, "gnu-emacs",                   BL, false,          boolPicks       ),
     //       new ParseString(TidyOptionId.eEmacsFile,             MS, "gnu-emacs-file",              ST, NullStr,        NullArr         ),
            new ParseAutoBool(TidyOptionId.eLiteralAttribs,      MU, "literal-attributes",          BL, TidyTriState.eNo,  boolPicks       ),

            new ParseBool(TidyOptionId.eForceOutput,             MS, "force-output",                BL, false,          boolPicks       ),
#if false
            new ParseBool(TidyOptionId.eBodyOnly,                MU, "show-body-only",              IN, false,          autoBoolPicks   ),
            new ParseBool(TidyOptionId.eFixUri,                  MU, "fix-uri",                     BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eLowerLiterals,           MU, "lower-literals",              BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eHideComments,            MU, "hide-comments",               BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eIndentCdata,             MU, "indent-cdata",                BL, false,          boolPicks       ),
           
            new ParseInt(TidyOptionId.eShowErrors,               DG, "show-errors",                 IN, 6,              NullArr         ),
            new ParseBool(TidyOptionId.eAsciiChars,              CE, "ascii-chars",                 BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eJoinClasses,             MU, "join-classes",                BL, false,          boolPicks       ),
            new ParseBool(TidyOptionId.eJoinStyles,              MU, "join-styles",                 BL, true,           boolPicks       ),
            new ParseBool(TidyOptionId.eEscapeCdata,             MU, "escape-cdata",                BL, false,          boolPicks       ),
#if SUPPORT_ASIAN_ENCODINGS
            new ParseBool(TidyOptionId.eLanguage,                CE, "language",                    ST, 0,             ParseName,         NULL            ),
            new ParseBool(TidyOptionId.eNCR,                     MU, "ncr",                         BL, true,          ParseBool,         boolPicks       ),
#endif
#if SUPPORT_UTF16_ENCODINGS
            new ParseBool(TidyOptionId.eOutputBOM,               CE, "output-bom",                  IN, TidyAutoState,   ParseAutoBool,     autoBoolPicks   ),
#endif
            new ParseBool(TidyOptionId.eReplaceColor,            MU, "replace-color",               BL, false,         boolPicks       ),
            new ParseCSS1Selector(TidyOptionId.eCSSPrefix,       MU, "css-prefix",                  ST, NullStr,       NullArr         ),
            new ParseTagNames(TidyOptionId.eInlineTags,          MU, "new-inline-tags",             ST, NullStr,       NullArr         ),
            new ParseTagNames(TidyOptionId.eBlockTags,           MU, "new-blocklevel-tags",         ST, NullStr,       NullArr         ),
            new ParseTagNames(TidyOptionId.eEmptyTags,           MU, "new-empty-tags",              ST, NullStr,       NullArr         ),
            new ParseTagNames(TidyOptionId.ePreTags,             MU, "new-pre-tags",                ST, NullStr,       NullArr         ),
//         new ParseAcc(TidyOptionId.eAccessibilityCheckLevel,  DG, "accessibility-check",         IN, NullStr,       accessPicks     ),
            new ParseBool(TidyOptionId.eVertSpace,               PP, "vertical-space",              BL, false,         boolPicks       ),
#if SUPPORT_ASIAN_ENCODINGS
            new ParseBool(TidyOptionId.ePunctWrap,               PP, "punctuation-wrap",            BL, false,              ParseBool,         boolPicks       ),
#endif
            new ParseBool(TidyOptionId.eMergeEmphasis,          MU, "merge-emphasis",              BL, true,                boolPicks       ),
            new ParseAutoBool(TidyOptionId.eMergeDivs,          MU, "merge-divs",                  IN, TidyTriState.eAuto,  autoBoolPicks   ),
            new ParseBool(TidyOptionId.eDecorateInferredUL,     MU, "decorate-inferred-ul",        BL, false,               boolPicks       ),
            new ParseBool(TidyOptionId.ePreserveEntities,       MU, "preserve-entities",           BL, false,               boolPicks       ),
            new ParseSorter(TidyOptionId.eSortAttributes,       PP, "sort-attributes",             IN, TidyAttrSortStrategy.eNone,   sorterPicks     ),
            new ParseAutoBool(TidyOptionId.eMergeSpans,         MU, "merge-spans",                 IN, TidyTriState.eAuto,  autoBoolPicks   ),
            new ParseBool(TidyOptionId.eAnchorAsName,           MU, "anchor-as-name",              BL, true,                boolPicks       ),
#endif
            };

            if (sOptAndPropDic == null)
            {
                sOptToIdDic = new Dictionary<string, TidyOptionId>();
                sOptAndPropDic = new Dictionary<TidyOptionId, OptAndProp>();

                string[] tidyOptionIdNames = Enum.GetNames(typeof(TidyOptionId));
                PropertyInfo[] optionsPropertyInfos = typeof(TidyOptions).GetProperties(BindingFlags.Public | BindingFlags.Instance);
                foreach (TidyOptionImpl opt in sOptionDefaults)
                {
                    sOptToIdDic.Add(opt.m_name, opt.m_id);
                    string optIdName = Enum.GetName(typeof(TidyOptionId), opt.m_id).Substring(1);   // remove leading 'e'.
                    PropertyInfo propInfo = Array.Find<PropertyInfo>(optionsPropertyInfos, 
                        o => o.Name.Equals(optIdName));

                    if (propInfo != null)
                        sOptAndPropDic.Add(opt.m_id, new OptAndProp() { mOptionImpl = opt, mPropertyInfo = propInfo });
                    else
                        Console.Error.WriteLine("No enum for e" + optIdName);
                
                }
            }
        }

        public static string GetOptionName(TidyOptionId optionId)
        {
            return TidyConfig.sOptAndPropDic[TidyOptionId.eInCharEncoding].mOptionImpl.m_name;
        }

        public bool Parse(Tidy tidy, string optName, string value)
        {
            TidyOptionId optionId;
            bool okay = sOptToIdDic.TryGetValue(optName.ToLower(), out optionId) && 
                sOptAndPropDic[optionId].mOptionImpl.Parse(tidy, value);
            return okay;
        }

        public bool GetPossibleValues(string optName, ref string errMsg)
        {
            TidyOptionId optionId;
            bool okay = sOptToIdDic.TryGetValue(optName.ToLower(), out optionId);
            if (okay)
            {
                string[] pickList = sOptAndPropDic[optionId].mOptionImpl.m_pickList;
                errMsg += "Possible options for command:" + optName + "\n";
                foreach (string pickStr in pickList)
                    errMsg += "  " + pickStr + "\n";
            }

            return okay;
        }

        public void LoadDefault(Tidy tidy)
        {
            foreach (TidyOptionImpl optionImpl in sOptionDefaults)
            {
                optionImpl.Parse(tidy, NullStr);    // Passing null forces default to be used.
            }
        }

        public void ShowConfig(Tidy tidy, TextWriter writer)
        {
            writer.WriteLine(string.Format("# {0} {1:g}", "Tidy Config on ", DateTime.Now));
            // foreach (TidyOptionId key in sGetSetMethodinfoDic.Keys)
            foreach (TidyOptionImpl opt in sOptionDefaults)
            {
                if (sOptAndPropDic.ContainsKey(opt.m_id))
                {
                    PropertyInfo propInfo = sOptAndPropDic[opt.m_id].mPropertyInfo;
                    object obj = opt.Get(tidy.Options);
                    if (obj != null)
                        writer.WriteLine(string.Format("{0} : {1}", opt.m_name.PadLeft(20), obj.ToString()));
                    else
                        writer.WriteLine(string.Format("{0}", opt.m_name.PadLeft(20)));
                }
            }
            writer.WriteLine("");
        }

        public bool LoadConfig(Tidy tidy, string cfgFile, out string errMsg)
        {
            int errCnt = 0;
            int lineCnt = 0;
            errMsg = "";
            char[] sSep = new char[] { ':', '=' };

            try
            {
                TextReader textReader = new StreamReader(cfgFile);
                string cfgLine;
                while ((cfgLine = textReader.ReadLine()) != null)
                {
                    if (cfgLine[0] == '#')
                        continue;

                    lineCnt++;
                    string[] tokenValue = cfgLine.Split(sSep, 2, StringSplitOptions.None);
                
                    TidyOptionId optionId;
                    if (sOptToIdDic.TryGetValue(tokenValue[0].Trim().ToLower(), out optionId))
                    {
                        TidyOptionImpl optionImpl = sOptAndPropDic[optionId].mOptionImpl;
                        string value = (tokenValue.Length == 2) ? tokenValue[1] : null;
                        if (value.Length > 1)
                            value = value.Trim();
                        if (!optionImpl.Parse(tidy, value))
                        {
                            if (errCnt == 0)
                                errMsg += string.Format("Error loading configuration file: {0}\n", cfgFile);
                            errMsg += string.Format("Line {0}, Don't understand:{1}\n", lineCnt, cfgLine);
                            errCnt++;
                        }
                    }
                }

                textReader.Close();
            } 
            catch (Exception ex)
            {
                errCnt++;
                errMsg += ex.Message + "\n";
            }
            
             
            return (errCnt == 0);
        }

        public bool SaveConfig(Tidy tidy, string cfgFile, out string errMsg)
        {
            errMsg = "";

            try
            {
                TextWriter textWriter = new StreamWriter(cfgFile);
                textWriter.WriteLine("# " + DateTime.Now.ToString("g"));

                foreach (TidyOptionImpl optionImpl in sOptionDefaults)
                {
                    string curValue = optionImpl.Get(tidy.Options).ToString();
                    if (!curValue.Equals(optionImpl.DefaultStr))
                    {
                        textWriter.WriteLine(string.Format("{0}: {1}", optionImpl.m_name, curValue));
                    }
                }
                textWriter.Close();
            }
            catch (Exception ex)
            {
                errMsg += ex.Message + "\n";
            }

            return !string.IsNullOrEmpty(errMsg);
        }
    }
}
