﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Net;
using System.Diagnostics;

namespace TidyCmd
{
    /// <summary>
    /// Classes to access web documents, part of TidyCmd.
    /// 
    /// Created by Dennis Lang - 2014
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// ========== License ==========
    /// This file is part of TidyCmd.
    ///
    /// TidyCmd is free software: you can redistribute it and/or modify
    /// it under the terms of the GNU General Public License as published by
    /// the Free Software Foundation, either version 3 of the License, or
    /// (at your option) any later version.
    ///
    /// TidyCmd is distributed in the hope that it will be useful,
    /// but WITHOUT ANY WARRANTY; without even the implied warranty of
    /// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    /// GNU General Public License for more details.
    ///
    /// The GNU General Public License: <http://www.gnu.org/licenses/>.
    /// </summary>
    class WebCmd
    {
        /// <summary>
        /// Holds information on Web Document access. 
        /// </summary>
        public class WebInfo
        {
            public bool mValid = false;
            // public Uri mUri = null;
            public string mPath = string.Empty;
            public long contentLength = 0;
            public string contentType = string.Empty;
            public DateTime lastModified = DateTime.MinValue;
            public string server = string.Empty;
            public TimeSpan loadTime = TimeSpan.FromSeconds(0);

            public static implicit operator bool(WebInfo webInfo)
            {
                return (webInfo == null) ? false : webInfo.mValid;
            }

            public string Summary()
            {
                string s1 = string.Format("Length:{0:N0} Type:{1} ", contentLength, contentType);
                string s2 = string.Empty;
                if (!String.IsNullOrEmpty(server) || lastModified != DateTime.MinValue)
                    s2 = string.Format("Server:{0} Last-Modified:{1:g} LoadSeconds:{2:N1}",
                        server, lastModified, loadTime.Milliseconds / 1000.0);
                return s1 + s2;
            }
        }

        public delegate int ProcData(Stream inStr, string path);

        public static  string MissingHostName = "Missing host name";
        public static  string HostChanged = "Host changed";

        /// <summary>
        /// Make a Uri from a Base Uri and Relative or Absolute path.
        /// </summary>
        /// <param name="baseUri"></param>
        /// <param name="filePath"></param>
        /// <returns></returns>
        protected Uri MakeUri(Uri baseUri, string filePath)
        {
            Uri fileUri = new Uri(baseUri, filePath);

            // if (!fileUri.IsAbsoluteUri)
            //     fileUri = new Uri(baseUri, fileUri);
            return fileUri;
        }

        /// <summary>
        /// Determine if Base+File Uri points to an existing path.
        /// If Mode is set to 'host' return HostChanged if host name differs from base. 
        /// </summary>
        /// <param name="baseUri"></param>
        /// <param name="fileUri"></param>
        /// <param name="mode"></param>
        /// <param name="msg"></param>
        /// <returns></returns>
        public virtual bool ValidPath(Uri baseUri, Uri fileUri, string mode, ref string msg)
        {
            if (fileUri.Host.Length != 0)
            {
                if (!string.IsNullOrEmpty(mode) && mode.StartsWith("host") && !fileUri.Host.Equals(baseUri.Host))
                {
                    msg = HostChanged;
                    return false;
                }
            }
            else
            {
                if (fileUri.IsFile)
                {
                    string ext = Path.GetExtension(fileUri.AbsolutePath).ToLower();
                    string validExtensions = ".htm|.html|.xhtml";
                    return validExtensions.Contains(ext);
                }
                msg = MissingHostName;
                return false;
            }
            return true;
        }

        /// <summary>
        /// Get page data and execute delegate ProcData to process data.
        /// </summary>
        /// <param name="baseUri"></param>
        /// <param name="fileUri"></param>
        /// <param name="msg"></param>
        /// <param name="mode"></param>
        /// <param name="urlType"></param>
        /// <param name="procData"></param>
        /// <returns></returns>
        public WebInfo GetData(Uri baseUri, Uri fileUri, ref string msg, string mode, string urlType, ProcData procData)
        {
            WebInfo webInfo = new WebInfo() { mPath = fileUri.ToString() };

            try
            {
                if (ValidPath(baseUri, fileUri, mode, ref msg))
                {
                    HttpWebRequest request = WebRequest.Create(fileUri) as HttpWebRequest;
                    request.UserAgent = "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.71 Safari/537.36";
               //     request.AllowAutoRedirect = false; 
               //     request.MaximumAutomaticRedirections = 1;
                    request.Timeout = 5000;
                    request.ServicePoint.ConnectionLeaseTimeout = 5000;
                    request.ServicePoint.MaxIdleTime = 5000;

                    Stopwatch stopwatch = new Stopwatch();
                    stopwatch.Start();
                    HttpWebResponse response;
                    response = request.GetResponse() as HttpWebResponse;

                    webInfo.loadTime = stopwatch.Elapsed;
                    webInfo.contentType = response.ContentType;
                    webInfo.contentLength = response.ContentLength;
                    webInfo.lastModified = response.LastModified;
                    webInfo.server = response.Server;
                    webInfo.mValid = response.StatusCode == HttpStatusCode.OK
                        && request.HaveResponse
                        && (string.IsNullOrEmpty(urlType) || response.ContentType.StartsWith(urlType));

                    if (webInfo.mValid && request.AllowAutoRedirect && !response.ResponseUri.Equals(fileUri))
                    {
                        // Try and detect ISP Domain Helper service mapping a failed page load to a 
                        // user friendly error presentation page. 
                        // Comcast example: comcast.com/site/error/error_xxx.html
                        // See xfinity DNSSEC is enhanced security
                        if (response.ResponseUri.GetLeftPart(UriPartial.Path).ToLower().Contains("error"))
                        {
                            webInfo.mValid = false;
                            msg = response.ResponseUri.GetLeftPart(UriPartial.Path);
                        }
                        else
                            msg = "Redirected to " + response.ResponseUri.AbsoluteUri;
                    }

                    if (webInfo.mValid && procData != null)
                    {
                        Console.Error.WriteLine(response.ContentType + " " + fileUri);      // #### DEBUG ####
                        procData(response.GetResponseStream(), fileUri.ToString());
                    }
                    else if (response.StatusCode != HttpStatusCode.OK)
                        msg = response.StatusDescription;

                    response.Close();
                }
            }
            catch (Exception ex)
            {
                msg = ex.Message;
            }

            return webInfo;
        }
    }

    /// <summary>
    /// Specialize version of WebCmd which caches page validation/load results and avoids
    /// redundant loading. 
    /// </summary>
    class WebPageCmd : WebCmd
    {
        public Dictionary<string, WebInfo> visitedPages = new Dictionary<string, WebInfo>();
        public static string IgnoredDuplicate = "Duplicate";

        /// <summary>
        /// Return True if Uri points to an existing file or Web resource. 
        /// </summary>
        /// <param name="baseUri"></param>
        /// <param name="fileUri"></param>
        /// <param name="mode"></param>
        /// <param name="msg"></param>
        /// <returns></returns>
        public override bool ValidPath(Uri baseUri, Uri fileUri, string mode, ref string msg)
        {
            bool okay = base.ValidPath(baseUri, fileUri, mode, ref msg);
            if (okay)
            {
                if (visitedPages.ContainsKey(fileUri.ToString()))
                {
                    msg = IgnoredDuplicate;
                    return false;
                    // return visitedPages[fileUri.ToString()];
                }
                visitedPages.Add(fileUri.ToString(), null);
            }
            return okay;
        }

        /// <summary>
        /// Return True if path points to an existing file or web resource. 
        /// </summary>
        /// <param name="basePath"></param>
        /// <param name="urlPath"></param>
        /// <param name="mode"></param>
        /// <param name="msg"></param>
        /// <returns></returns>
        public bool ValidPath(string basePath, string urlPath, string mode, ref string msg)
        {
            bool okay = false;
            try
            {
                Uri baseUri = new Uri(basePath);
                Uri fileUri = MakeUri(baseUri, urlPath);
                okay = ValidPath(baseUri, fileUri, mode, ref msg);
            }
            catch (UriFormatException uex)
            {
                msg = uex.Message;
            }
            catch (Exception ex)
            {
                msg = ex.Message;
            }
            return okay;
        }

        /// <summary>
        /// Load Web Page and execute delegate ProcData to process it.
        /// </summary>
        /// <param name="basePath"></param>
        /// <param name="urlPath"></param>
        /// <param name="msg"></param>
        /// <param name="mode"></param>
        /// <param name="urlType"></param>
        /// <param name="procData"></param>
        /// <returns></returns>
        public WebInfo LoadPage(string basePath, string urlPath, ref string msg, string mode, string urlType, ProcData procData)
        {
            WebInfo webInfo = null;
            try
            {
                Uri baseUri = new Uri(basePath);
                Uri fileUri = MakeUri(baseUri, urlPath);

                webInfo = GetData(baseUri, fileUri, ref msg, mode, urlType, procData);
                if (!msg.Equals(IgnoredDuplicate))
                    visitedPages[fileUri.ToString()] = webInfo;
                else
                    webInfo = visitedPages[fileUri.ToString()];
            }
            catch (Exception ex)
            {
                msg = ex.Message;
            }

            return webInfo;
        }
    }
 
}
