using System;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using TidyNetLib.tidy;

namespace TidyNetLib
{
	/// <summary>
	/// Tidy options.
	/// 
	/// (c) 1998-2000 (W3C) MIT, INRIA, Keio University
	/// See Tidy.cs for the copyright notice.
	/// Derived from <a href="http://www.w3.org/People/Raggett/tidy">
	/// HTML Tidy Release 4 Aug 2000</a>
	/// 
	/// </summary>
	/// <author>Dave Raggett &lt;dsr@w3.org&gt;</author>
	/// <author>Andy Quick &lt;ac.quick@sympatico.ca&gt; (translation to Java)</author>
	/// <author>Seth Yates &lt;seth_yates@hotmail.com&gt; (translation to C#)</author>
	/// <version>1.0, 1999/05/22</version>
	/// <version>1.0.1, 1999/05/29</version>
	/// <version>1.1, 1999/06/18 Java Bean</version>
	/// <version>1.2, 1999/07/10 Tidy Release 7 Jul 1999</version>
	/// <version>1.3, 1999/07/30 Tidy Release 26 Jul 1999</version>
	/// <version>1.4, 1999/09/04 DOM support</version>
	/// <version>1.5, 1999/10/23 Tidy Release 27 Sep 1999</version>
	/// <version>1.6, 1999/11/01 Tidy Release 22 Oct 1999</version>
	/// <version>1.7, 1999/12/06 Tidy Release 30 Nov 1999</version>
	/// <version>1.8, 2000/01/22 Tidy Release 13 Jan 2000</version>
	/// <version>1.9, 2000/06/03 Tidy Release 30 Apr 2000</version>
	/// <version>1.10, 2000/07/22 Tidy Release 8 Jul 2000</version>
	/// <version>1.11, 2000/08/16 Tidy Release 4 Aug 2000</version>
	public class TidyOptions
	{
		/// <summary>
		/// Default constructor.
		/// </summary>
		/// <param name="configuration"></param>
        public TidyOptions()
        {
            IndentSpaces = 2;                       // default indentation 
            WrapLen = 68;                           // default wrap margin 
            InCharEncoding = TidyNetLib.CharEncoding.ASCII;
            OutCharEncoding = TidyNetLib.CharEncoding.ASCII;
            TabSize = 4;

            AltText = null;                         // default text for alt attribute 
            SlideStyle = null;                      // style sheet for slides 
            DocType = TidyNetLib.DocType.Auto;      // user specified doctype 

            IndentContent = false;                  // indent content of appropriate tags 
            SmartIndent = false;                    // does text/block level content effect indentation 
            HideEndTags = false;                    // suppress optional end tags 
            XmlTags = false;                        // treat input as XML 
            XmlOut = false;                         // create output as XML 
            XhtmlOut = false;                       // output extensible HTML 
            XmlDecl = false;                        // add <?xml?> for XML docs 
            RawOut = false;                         // avoid mapping values > 127 to entities 
            UpperCaseTags = false;                  // output tags in upper not lower case 
            UpperCaseAttrs = false;                 // output attributes in upper not lower case 
            MakeClean = false;                      // remove presentational clutter 
            LogicalEmphasis = false;                // replace i by em and b by strong 
            DropFontTags = false;                   // discard presentation tags 
            DropEmptyParas = true;                  // discard empty p elements 
            FixComments = true;                     // fix comments with adjacent hyphens 
            BreakBeforeBR = false;                  // o/p newline before <br> or not? 
            BurstSlides = false;                    // create slides on each h2 element 
            NumEntities = false;                    // use numeric entities 
            QuoteMarks = false;                     // output " marks as &quot; 
            QuoteNbsp = true;                       // output non-breaking space as entity 
            QuoteAmpersand = true;                  // output naked ampersand as &amp; 
            WrapAttVals = false;                    // wrap within attribute values 
            WrapScriptlets = false;                 // wrap within JavaScript string literals 
            WrapSection = true;                     // wrap within <![ ... ]> section tags 
            WrapAsp = true;                         // wrap within ASP pseudo elements 
            WrapJste = true;                        // wrap within JSTE pseudo elements 
            WrapPhp = true;                         // wrap within PHP pseudo elements 
            FixBackslash = true;                    // fix URLs by replacing \ with / 
            IndentAttributes = false;               // newline+indent before each attribute 
            XmlPIs = false;                         // if set to yes PIs must end with ?> 
            XmlSpace = false;                       // if set to yes adds xml:space attr as needed 
            EncloseBodyText = false;                // if yes text at body is wrapped in <p>'s 
            EncloseBlockText = false;               // if yes text in blocks is wrapped in <p>'s 
            Word2000 = false;                       // draconian cleaning for Word2000 
            TidyMark = true;                        // add meta element indicating tidied doc 
            LiteralAttribs = false;                 // if true attributes may use newlines 
            DuplicateAttrs = true;                  // Remove duplicate attributes
            tt = new TagTable();
            Styles = new StyleList();
            AHrefSet = new HashSet<string>();
            ANameSet = new HashSet<string>();
            ImgSrcSet = new HashSet<string>();
            
#if false
            styleTbl = new StyleTable();
#endif

            ForceOutput = false;                     // If true output even if errors.
        }

        /// <summary>
        /// Clear temporary lists maintained to keep stats on page loaded. 
        /// Stats are per page and should be cleared before loading a new page.
        /// </summary>
        public void Clear()
        {
            Styles.Clear();
            AHrefSet.Clear();
            ANameSet.Clear();
            ImgSrcSet.Clear();
        }

        /// <summary>
        /// Wrapper to store Styles along with a count of their Explicit references. 
        /// </summary>
        public class StyleList : SortedDictionary<string, int> 
        {
            int Find(Regex findThis)
            {
                int idx = 0;
                foreach (string style in this.Keys)
                {
                    if (findThis.IsMatch(style))
                    {
                        this[style] = this[style] + 1;
                        return idx;
                    }
                    idx++;
                }
                return -1;
            }

            public bool Add(string styleTag)
            {
                if (this.ContainsKey(styleTag))
                    return false;   // duplicate style
                this.Add(styleTag, 0);
                return true;
            }

            public bool ContainsClass(string element, string value)
            {
                return Find(new Regex(string.Format(@"({0}|)\.{1}(:[a-z]|)", element, value), RegexOptions.IgnoreCase)) != -1;
            }

            public bool ContainsId(string element, string value)
            {
                return Find(new Regex(string.Format(@"({0}|)\#{1}(:[a-z]|)", element, value), RegexOptions.IgnoreCase)) != -1;
            }
        }

		/// <summary>Default indentation</summary>
		public virtual int IndentSpaces { get; set; }

		/// <summary>Default wrap margin</summary>
		public virtual int WrapLen { get; set; }

		/// <summary>Character Encoding</summary>
        public virtual CharEncoding OutCharEncoding { get; set; }
        public virtual CharEncoding InCharEncoding { get; set; }
        public virtual CharEncoding CharEncoding
        { set { OutCharEncoding = InCharEncoding = value; } get { return OutCharEncoding; } }

		/// <summary>Tab size</summary>
		public virtual int TabSize { get; set; }

		/// <summary>Indent content of appropriate tags</summary>
        public virtual bool IndentContent { get; set; }

		/// <summary>Does text/block level content affect indentation</summary>
        public virtual bool SmartIndent { get; set; }

		/// <summary>Suppress optional end tags</summary>
        public virtual bool HideEndTags { get; set; }

		/// <summary>Treat input as XML</summary>
        public virtual bool XmlTags { get; set; }

		/// <summary>Create output as XML</summary>
        public virtual bool XmlOut  { get;  set; }

		/// <summary>Output XHTML</summary>
		public virtual bool XhtmlOut { get; set; }
		
		/// <summary>Avoid mapping values > 127 to entities</summary>
		public virtual bool RawOut { get; set; }

		/// <summary>Output tags in upper not lower case</summary>
		public virtual bool UpperCaseTags { get; set; }

		/// <summary>Output attributes in upper not lower case</summary>
		public virtual bool UpperCaseAttrs { get; set; }

		/// <summary>Remove presentational clutter</summary>
        public virtual bool MakeClean { get; set; }

		/// <summary>O/p newline before &lt;br&gt; or not?</summary>
		public virtual bool BreakBeforeBR { get; set; }

		/// <summary>Create slides on each h2 element</summary>
		public virtual bool BurstSlides { get; set; }

		/// <summary>Use numeric entities</summary>
		public virtual bool NumEntities { get; set; }

		/// <summary>Output " marks as &amp;quot;</summary>
		public virtual bool QuoteMarks { get; set; }

		/// <summary>Output non-breaking space as entity</summary>
		public virtual bool QuoteNbsp { get; set; }

		/// <summary>Output naked ampersand as &amp;</summary>
		public virtual bool QuoteAmpersand { get; set; }

		/// <summary>Wrap within attribute values</summary>
		public virtual bool WrapAttVals { get; set; }

		/// <summary>Wrap within JavaScript string literals</summary>
		public virtual bool WrapScriptlets { get; set; }

		/// <summary>Wrap within &lt;![ ... ]&gt; section tags</summary>
		public virtual bool WrapSection { get; set; }

		/// <summary>Default text for alt attribute</summary>
        public virtual string AltText { get; set; }

		/// <summary>Style sheet for slides</summary>
		public virtual string SlideStyle { get; set; }

        /// <summary>Add <?XML version="1.0"?>  for XML docs</summary>
		public virtual bool XmlDecl { get; set; }   // xmlPi

		/// <summary>Discard presentation tags</summary>
        public virtual bool DropFontTags { get; set; }

		/// <summary>Discard empty p elements</summary>
		public virtual bool DropEmptyParas { get; set; }

		/// <summary>Fix comments with adjacent hyphens</summary>
		public virtual bool FixComments { get; set; }

		/// <summary>Wrap within ASP pseudo elements</summary>
		public virtual bool WrapAsp { get; set; }

		/// <summary>Wrap within JSTE pseudo elements</summary>
		public virtual bool WrapJste { get; set; }

		/// <summary>Wrap within PHP pseudo elements</summary>
		public virtual bool WrapPhp { get; set; }

		/// <summary>Fix URLs by replacing \ with /</summary>
		public virtual bool FixBackslash { get; set; }

		/// <summary>Newline+indent before each attribute</summary>
		public virtual bool IndentAttributes { get; set; }

		/// <summary>Replace i by em and b by strong</summary>
		public virtual bool LogicalEmphasis { get; set; }

		/// <summary>If set to true PIs must end with ?></summary>
		public virtual bool XmlPIs { get; set; }

		/// <summary>If true text at body is wrapped in &lt;p&gt;'s</summary>
        public virtual bool EncloseBodyText { get; set; }   

		/// <summary>If true text in blocks is wrapped in &lt;p&gt;'s</summary>
		public virtual bool EncloseBlockText { get; set; }

		/// <summary>Draconian cleaning for Word2000</summary>
		public virtual bool Word2000 { get; set; }

		/// <summary>Add meta element indicating tidied doc</summary>
		public virtual bool TidyMark { get; set; }

		/// <summary>If set to yes adds xml:space attr as needed</summary>
		public virtual bool XmlSpace { get; set; }

		/// <summary>If true attributes may use newlines</summary>
		public virtual bool LiteralAttribs { get; set; }

        /// <summary>If true remove duplicate attributes</summary>
        public virtual bool DuplicateAttrs { get; set; }

        /// <summary> Output even if there are errors </summary>
        public virtual bool ForceOutput { get; set; }

        /// <summary>  Set of HRef found in document </summary>
        public HashSet<string> AHrefSet { get; set; }
        /// <summary>  Set of A-tag Names found in document </summary>
        public HashSet<string> ANameSet { get; set; }
        /// <summary> Set of Images found in doucment></summary>
        public HashSet<string> ImgSrcSet { get; set; }

		/// <summary>
		/// The DOCTYPE
		/// </summary>
		public virtual DocType DocType { get; set; }

		internal class Tokenizer
		{
			public Tokenizer(string source, string delimiters)
			{
				_elements = new ArrayList();
				_delimiters = delimiters;
				_elements.AddRange(source.Split(_delimiters.ToCharArray()));
				for (int index = 0; index < _elements.Count; index++)
				{
					if ((string)_elements[index] == "")
					{
						_elements.RemoveAt(index);
						index--;
					}
				}
				_source = source;
			}

			public bool HasMoreTokens()
			{
				return (_elements.Count > 0);			
			}

			public string NextToken()
			{			
				string result;
				if (_source.Length == 0)
				{
					throw new Exception();
				}
				else
				{
					_elements = new ArrayList();
					_elements.AddRange(_source.Split(_delimiters.ToCharArray()));
					for (int index = 0; index < _elements.Count; index++)
					{
						if ((string)_elements[index] == "")
						{
							_elements.RemoveAt(index);
							index--;
						}
					}
					result = (string) _elements[0];
					_elements.RemoveAt(0);				
					_source = _source.Remove(_source.IndexOf(result), result.Length);
					_source = _source.TrimStart(_delimiters.ToCharArray());
					return result;					
				}			
			}

			private ArrayList _elements;
			private string _source;
			private string _delimiters = " \t\n\r";
		}

		internal virtual TagTable tt { get; set; }

        public StyleList Styles { get; set; }

		/* ensure that config is self consistent */
		internal virtual void Adjust()
		{
			if (EncloseBlockText)
				EncloseBodyText = true;
			
			/* avoid the need to set IndentContent when SmartIndent is set */
			
			if (SmartIndent)
				IndentContent = true;
			
			/* disable wrapping */
			if (WrapLen == 0)
				WrapLen = 0x7FFFFFFF;
			
			/* Word 2000 needs o:p to be declared as inline */
			if (Word2000)
				tt.DefineInlineTag("o:p");
			
			/* XHTML is written in lower case */
			if (XhtmlOut)
			{
				XmlOut = true;
				UpperCaseTags = false;
				UpperCaseAttrs = false;
			}
			
			/* if XML in, then XML out */
			if (XmlTags)
			{
				XmlOut = true;
				XmlPIs = true;
			}
			
			/* XML requires end tags */
			if (XmlOut)
			{
				QuoteAmpersand = true;
				HideEndTags = false;
			}
		}
		
		private void ParseInlineTagNames(string s, string option)
		{
			Tokenizer t = new Tokenizer(s, " \t\n\r,");
			while (t.HasMoreTokens())
			{
				tt.DefineInlineTag(t.NextToken());
			}
		}
		
		private void ParseBlockTagNames(string s, string option)
		{
			Tokenizer t = new Tokenizer(s, " \t\n\r,");
			while (t.HasMoreTokens())
			{
				tt.DefineBlockTag(t.NextToken());
			}
		}
		
		private void ParseEmptyTagNames(string s, string option)
		{
			Tokenizer t = new Tokenizer(s, " \t\n\r,");
			while (t.HasMoreTokens())
			{
				tt.defineEmptyTag(t.NextToken());
			}
		}
		
		private void ParsePreTagNames(string s, string option)
		{
			Tokenizer t = new Tokenizer(s, " \t\n\r,");
			while (t.HasMoreTokens())
			{
				tt.DefinePreTag(t.NextToken());
			}
		}
		

	}
}
